/************************************************************************
    MeOS - Orienteering Software
    Copyright (C) 2009-2010 Melin Software HB
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Melin Software HB - software@melin.nu - www.melin.nu
    Stigbergsvgen 7, SE-75242 UPPSALA, Sweden
    
************************************************************************/

#include "stdafx.h"

#include "resource.h"

#include <commctrl.h>
#include <commdlg.h> 

#include "oEvent.h"
#include "xmlparser.h"
#include "gdioutput.h"
#include "csvparser.h"
#include "SportIdent.h"
#include "meos_util.h"
#include "TabCompetition.h"
#include "oFreeImport.h"
#include "localizer.h"
#include "oListInfo.h"

#include <cassert>
#include <cmath>

void Setup(bool overwrite);
void exportSetup();
void resetSaveTimer();

TabCompetition::TabCompetition(oEvent *poe):TabBase(poe)
{
  defaultServer="localhost";
  defaultName="meos";
}

TabCompetition::~TabCompetition(void)
{
}

extern SportIdent *gSI;
extern HINSTANCE hInst;
extern HWND hWndMain;

void runCourseImport(gdioutput& gdi, const string &filename, oEvent *oe) {
	csvparser csv;
  if (csv.iscsv(filename.c_str())) {
	  gdi.addString("", 0, "Importerar OCAD csv-fil...");
	  gdi.refresh();
		
	  if(csv.ImportOCAD_CSV(*oe, filename.c_str())) {
		  gdi.addString("", 0, "Klart.");
	  }
	  else gdi.addString("", 0, "Operationen misslyckades.");
  }
  else {
    oe->importXML_EntryData(gdi, filename.c_str());
  }

  gdi.setWindowTitle(oe->getTitleName());
  oe->updateTabs();
	gdi.refresh();
}

bool TabCompetition::save(gdioutput &gdi, bool write)
{
	string name=gdi.getText("Name");

  if(name.empty()) {
		gdi.alert("Tvlingen mste ha ett namn");
		return 0;
	}

	oe->setName(gdi.getText("Name"));
	oe->setDate(gdi.getText("Date"));
	oe->setZeroTime(gdi.getText("ZeroTime"));
  
  oe->synchronize();
	if(gSI) gSI->SetZeroTime(oe->getZeroTimeNum());

	gdi.setWindowTitle(oe->getTitleName());

  if (write) {
	  gdi.setWaitCursor(true);
    resetSaveTimer();
	  return oe->save();
  }
  else
    return true;
}

string TabCompetition::browseForOpen(const string &Filter, const string &defext)
{
	InitCommonControls();
	
	char FileName[260];
	FileName[0]=0;
	char sbuff[256];
	OPENFILENAME of;
	sprintf_s(sbuff, 256, "%s", Filter.c_str());

	int sl=strlen(sbuff);
	for(int m=0;m<sl;m++) if(sbuff[m]=='') sbuff[m]=0;

	of.lStructSize      =sizeof(of);
	of.hwndOwner        =hWndMain;
	of.hInstance        =hInst;
	of.lpstrFilter      =sbuff;
	of.lpstrCustomFilter=NULL;
	of.nMaxCustFilter   =0;
	of.nFilterIndex     =1;
	of.lpstrFile        = FileName;
	of.nMaxFile         =260;
	of.lpstrFileTitle   =NULL;
	of.nMaxFileTitle    =0;
	of.lpstrInitialDir  =NULL;
	of.lpstrTitle       =NULL;
	of.Flags            =OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;
	of.lpstrDefExt   	  = defext.c_str();
	of.lpfnHook		     =NULL;
	
	
	if(GetOpenFileName(&of)==false)
		return "";
	
	return FileName;
}

string TabCompetition::browseForSave(const string &Filter, const string &defext, int &FilterIndex)
{
	InitCommonControls();
	
	char FileName[260];
	FileName[0]=0;
	char sbuff[256];
	OPENFILENAME of;
//	LoadString(hInst,1016,sbuff,256);
	sprintf_s(sbuff, 256,"%s", Filter.c_str());

	int sl=strlen(sbuff);
	for(int m=0;m<sl;m++) if(sbuff[m]=='') sbuff[m]=0;

	of.lStructSize      = sizeof(of);
	of.hwndOwner        = hWndMain;
	of.hInstance        = hInst;
	of.lpstrFilter      = sbuff;
	of.lpstrCustomFilter= NULL;
	of.nMaxCustFilter   = 0;
	of.nFilterIndex     = 1;
	of.lpstrFile        = FileName;
	of.nMaxFile         = 260;
	of.lpstrFileTitle   = NULL;
	of.nMaxFileTitle    = 0;
	of.lpstrInitialDir  = NULL;
	of.lpstrTitle       = NULL;
	of.Flags            = OFN_OVERWRITEPROMPT|OFN_HIDEREADONLY;
	of.lpstrDefExt   	  = defext.c_str();
	of.lpfnHook		      = NULL;
	
	if(GetSaveFileName(&of)==false)
		return "";

	FilterIndex=of.nFilterIndex;

	return FileName;
}

bool TabCompetition::importFile(HWND hWnd, gdioutput &gdi)
{
	InitCommonControls();
	
	char FileName[260];
	FileName[0]=0;
	char sbuff[256];
	OPENFILENAME of;
//	LoadString(hInst,1016,sbuff,256);
	sprintf_s(sbuff, 256, "XML-data*.xml;*.bu?");

	int sl=strlen(sbuff);
	for(int m=0;m<sl;m++) if(sbuff[m]=='') sbuff[m]=0;

	of.lStructSize      =sizeof of;
	of.hwndOwner        =hWnd;
	of.hInstance        =hInst;
	of.lpstrFilter      =sbuff;
	of.lpstrCustomFilter=NULL;
	of.nMaxCustFilter   =0;
	of.nFilterIndex     =1;
	of.lpstrFile        = FileName;
	of.nMaxFile         =260;
	of.lpstrFileTitle   =NULL;
	of.nMaxFileTitle    =0;
	of.lpstrInitialDir  =NULL;
	of.lpstrTitle       =NULL;
	of.Flags            =OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;
	of.lpstrDefExt   	  ="xml";
	of.lpfnHook		     =NULL;
	
	
	if(GetOpenFileName(&of)==false)
		return false;
	
	gdi.setWaitCursor(true);
	if(oe->open(FileName, true)) {
		if(gSI) gSI->SetZeroTime(oe->getZeroTimeNum());
		gdi.setWindowTitle(oe->getTitleName());
    resetSaveTimer();
		return true;
	}

	return false;
}

bool TabCompetition::exportFileAs(HWND hWnd, gdioutput &gdi)
{
	InitCommonControls();
	
	char FileName[260];
	FileName[0]=0;
	char sbuff[256];

	OPENFILENAME of;
	//LoadString(hInst,1016,sbuff,256);
	sprintf_s(sbuff, 256, "XML-data*.xml");
	
	int sl=strlen(sbuff);
	for(int m=0;m<sl;m++) if(sbuff[m]=='') sbuff[m]=0;

	of.lStructSize      =sizeof of;
	of.hwndOwner        =hWnd;
	of.hInstance        =hInst;
	of.lpstrFilter      =sbuff;
	of.lpstrCustomFilter=NULL;
	of.nMaxCustFilter   =0;
	of.nFilterIndex     =1;
	of.lpstrFile        = FileName;
	of.nMaxFile         = 260;
	of.lpstrFileTitle   = NULL;
	of.nMaxFileTitle    = 0;
	of.lpstrInitialDir  = NULL;
	of.lpstrTitle       = NULL;
	of.Flags            = OFN_OVERWRITEPROMPT|OFN_HIDEREADONLY;
	of.lpstrDefExt   	  = "xml";
	of.lpfnHook		      = NULL;
	
	
	if(GetSaveFileName(&of)==false)
		return false;

	gdi.setWaitCursor(true);
	if(!oe->save(FileName)) {
		gdi.alert("Fel: Filen "+string(FileName)+ " kunde inte skrivas.");
		return false;
	}

	return true;
}

int CompetitionCB(gdioutput *gdi, int type, void *data)
{
  TabCompetition &tc = dynamic_cast<TabCompetition &>(*gdi->getTabs().get(TCmpTab));

  return tc.competitionCB(*gdi, type, data);
}

int restoreCB(gdioutput *gdi, int type, void *data)
{
  TabCompetition &tc = dynamic_cast<TabCompetition &>(*gdi->getTabs().get(TCmpTab));

  return tc.restoreCB(*gdi, type, data);
}

void TabCompetition::loadConnectionPage(gdioutput &gdi)
{
  gdi.clearPage(false);
  showConnectionPage=true;    
  gdi.addString("", boldLarge, "Anslutningar");

  if(oe->getServerName().empty()) {
    gdi.addString("", 10, "help:52726");
    gdi.pushX();
    
    defaultServer = oe->getPropertyString("Server", defaultServer);
    defaultName = oe->getPropertyString("UserName", defaultName);
    defaultPort = oe->getPropertyString("Port", defaultPort);
    string client = oe->getPropertyString("Client", oe->getClientName());

    gdi.fillRight();
	  gdi.addInput("Server", defaultServer, 16, 0, "MySQL Server / IP-adress:", "IP-adress eller namn p en MySQL-server");
	  gdi.addInput("UserName", defaultName, 7, 0, "Anvndarnamn:");
	  gdi.addInput("PassWord", defaultPwd, 9, 0, "Lsenord:");
	  gdi.addInput("Port", defaultPort, 4, 0, "Port:");
    gdi.fillDown();
    gdi.popX();
    gdi.dropLine(2.5);
    gdi.addInput("ClientName", client, 16, 0, "Klientnamn");
    gdi.dropLine();
    gdi.fillRight();
    gdi.addButton("ConnectToMySQL", "Anslut", CompetitionCB);
  }
  else {
    gdi.addString("", 10, "help:50431");
    gdi.dropLine(1);
    gdi.pushX();
    gdi.fillRight();
    gdi.addString("", 1, "Ansluten till:");
    gdi.addStringUT(1, oe->getServerName()).setColor(colorGreen);
    gdi.popX();
    gdi.dropLine(2);
    gdi.addInput("ClientName", oe->getClientName(), 16, 0, "Klientnamn:");
    gdi.dropLine();
    gdi.addButton("SaveClient", "ndra", CompetitionCB);
    gdi.dropLine(2.5);

    gdi.popX();
    gdi.addString("", 1, "ppnad tvling:");
    
    if(oe->empty())
      gdi.addString("", 1, "Ingen").setColor(colorRed);
    else {
      gdi.addStringUT(1, oe->getName()).setColor(colorGreen);
    
      if(oe->isClient())
        gdi.addString("", 1, "(p server)");
      else
        gdi.addString("", 1, "(lokalt)");

    }
    gdi.dropLine(2);
    gdi.popX();
    gdi.fillRight();

    if(!oe->isClient())
      gdi.addButton("UploadCmp", "Ladda upp ppnad tvling p server",CompetitionCB);
   
    if(oe->empty())
      gdi.disableInput("UploadCmp");
    else {
      gdi.addButton("CloseCmp", "Stng tvlingen", CompetitionCB);
      gdi.addButton("Delete", "Radera tvlingen", CompetitionCB);
    }
    gdi.dropLine(2);
    gdi.popX();
    if (oe->empty()) {
		  char bf[260];
		  getUserFile(bf, "");
		  oe->enumerateCompetitions(bf, "*.meos");		
  		
      gdi.dropLine(1);
      gdi.fillRight();
      gdi.addListBox("ServerCmp", 320, 210,  CompetitionCB, "Server");
      oe->fillCompetitions(gdi, "ServerCmp", 2);
      gdi.selectItemByData("ServerCmp", oe->getPropertyInt("LastCompetition", 0));
      gdi.fillDown();   
      gdi.addListBox("LocalCmp", 320, 210, CompetitionCB, "Lokalt"); 
      gdi.popX();
      oe->fillCompetitions(gdi, "LocalCmp", 1);
      gdi.selectItemByData("LocalCmp", oe->getPropertyInt("LastCompetition", 0));
      
      gdi.fillRight();
      gdi.addButton("OpenCmp", "ppna tvling", CompetitionCB);    
    }
    else if (oe->isClient()) {
      gdi.fillDown();
      gdi.popX();
      oe->listConnectedClients(gdi);
      gdi.registerEvent("Connections", CompetitionCB);
      gdi.fillRight();
    }

    gdi.addButton("DisconnectMySQL", "Koppla ner databas", CompetitionCB);
  }
  gdi.addButton("Cancel", "Till huvudsidan", CompetitionCB);
  gdi.fillDown();
  gdi.refresh();
}

int TabCompetition::competitionCB(gdioutput &gdi, int type, void *data)
{
	if (type==GUI_BUTTON) {
		ButtonInfo bi=*(ButtonInfo *)data;

    if (bi.id=="UseEconomy" || bi.id=="UseSpeaker") {
      oe->getDI().setInt("UseEconomy", gdi.isChecked("UseEconomy"));
      oe->getDI().setInt("UseSpeaker", gdi.isChecked("UseSpeaker"));
      oe->updateTabs();
    }
    if (bi.id=="Print")	{
			gdi.print(oe);			
		}
    else if (bi.id=="Setup") {
      gdi.clearPage(false);

      gdi.addString("", boldLarge, "Instllningar MeOS");
      gdi.dropLine();
      gdi.addString("", 10, "help:29191");
      gdi.dropLine();
      char FileNamePath[260];
      getUserFile(FileNamePath, "");
      gdi.addStringUT(0, lang.tl("MeOS lokala datakatalog r") + ": " + FileNamePath);
      gdi.dropLine();

      char bf[260];
      GetCurrentDirectory(260, bf);
      gdi.fillRight();
      gdi.pushX();
      gdi.dropLine();
      gdi.addInput("Source", bf, 50, 0, "Kllkatalog:");
      gdi.dropLine(0.3);
      gdi.addButton("SourceBrowse", "Blddra...", CompetitionCB); 
      gdi.dropLine(4);
      gdi.popX();
      gdi.fillRight();
      gdi.addButton("DoSetup", "Installera", CompetitionCB);
      gdi.addButton("ExportSetup", "Exportera", CompetitionCB);
      gdi.addButton("Cancel", "Avbryt", CompetitionCB);
      gdi.dropLine(2);
      gdi.refresh();
    }
    else if (bi.id=="SourceBrowse") {
      string s = gdi.browseForFolder(gdi.getText("Source"));
      if (!s.empty())
        gdi.setText("Source", s);
    }
    else if (bi.id=="DoSetup") {
      string source = gdi.getText("Source");
      if(SetCurrentDirectory(source.c_str())) {
        Setup(true);
        gdi.alert("Tillgngliga filer installerades. Starta om MeOS.");
        exit(0);
      }
      else
        throw std::exception("Operationen misslyckades");
    }
    else if (bi.id=="ExportSetup") {
      gdi.clearPage(false);

      gdi.addString("", boldLarge, "Exportera instllningar och lpardatabaser");
      gdi.addString("", 10, "help:15491");
      gdi.dropLine();
      char FileNamePath[260];
      getUserFile(FileNamePath, "");
      gdi.addStringUT(0, lang.tl("MeOS lokala datakatalog r") + ": " + FileNamePath);
      
      gdi.dropLine();

      char bf[260];
      GetCurrentDirectory(260, bf);
      gdi.fillRight();
      gdi.pushX();
      gdi.dropLine();
      gdi.addInput("Source", bf, 50, 0, "Destinationskatalog:");
      gdi.dropLine(0.3);
      gdi.addButton("SourceBrowse", "Blddra...", CompetitionCB); 
      gdi.dropLine(4);
      gdi.popX();
      gdi.fillRight();
      gdi.addButton("DoExportSetup", "Exportera", CompetitionCB);
      gdi.addButton("Cancel", "Avbryt", CompetitionCB);
      gdi.dropLine(2);
      gdi.refresh();
    }
    else if (bi.id=="DoExportSetup") {
      string source = gdi.getText("Source");
      if(SetCurrentDirectory(source.c_str())) {
        exportSetup();
        gdi.alert("Instllningarna har exporterats.");
      }
    }
    else if (bi.id=="Report") {
      gdi.clearPage(true);
      oe->generateCompetitionReport(gdi);

      gdi.addButton(gdi.getWidth()+20, 15, gdi.scaleLength(120), "Cancel", 
                    "terg", CompetitionCB,  "", true, false);
      gdi.addButton(gdi.getWidth()+20, 18+gdi.getButtonHeight(), gdi.scaleLength(120), "Print", 
                    "Skriv ut...", CompetitionCB,  "Skriv ut rapporten", true, false);
      gdi.refresh();

      //gdi.addButton("Cancel", "Avbryt", CompetitionCB);    
    }
    else if (bi.id=="Settings") {
      gdi.clearPage(false);

      gdi.addString("", boldLarge, "Tvlingsinstllningar");
      gdi.dropLine(0.5);
      vector<string> fields;
      gdi.pushY();
      gdi.addString("", 1, "Adress och kontakt");
      fields.push_back("Organizer");
      fields.push_back("Street");
      fields.push_back("Address");
      fields.push_back("EMail");
      fields.push_back("Homepage");
      
      oe->getDI().buildDataFields(gdi, fields);
      
      gdi.newColumn();
      gdi.popY();

      gdi.addString("", 1, "Avgifter");
      fields.clear();
      gdi.fillRight();
      gdi.pushX();
      fields.push_back("CardFee");
      fields.push_back("EliteFee");
      fields.push_back("EntryFee");
      fields.push_back("YouthFee");

      oe->getDI().buildDataFields(gdi, fields);
      
      gdi.popX();
      gdi.dropLine(3);
      
      fields.clear();
      fields.push_back("OrdinaryEntry");
      fields.push_back("LateEntryFactor");

      oe->getDI().buildDataFields(gdi, fields);
      
      gdi.fillDown();
      gdi.popX();
      gdi.dropLine(3);
      
      gdi.addString("", 1, "Betalningsinformation");
      fields.clear();
      fields.push_back("Account");
      fields.push_back("PaymentDue");
      
      oe->getDI().buildDataFields(gdi, fields);
      oe->getDI().fillDataFields(gdi);

      gdi.fillRight();
      gdi.addButton("SaveSettings", "Spara", CompetitionCB);
      gdi.addButton("Cancel", "Avbryt", CompetitionCB);
      gdi.dropLine(2);
      gdi.refresh();
    }
    else if(bi.id=="SaveSettings") {
      oe->getDI().saveDataFields(gdi);
      if (oe->isChanged()) {
        oe->setProperty("Organizer", oe->getDCI().getString("Organizer"));
        oe->setProperty("Street", oe->getDCI().getString("Street"));
        oe->setProperty("Address", oe->getDCI().getString("Address"));
        oe->setProperty("EMail", oe->getDCI().getString("EMail"));
        oe->setProperty("Homepage", oe->getDCI().getString("Homepage"));

        oe->setProperty("CardFee", oe->getDCI().getInt("CardFee"));
        oe->setProperty("EliteFee", oe->getDCI().getInt("EliteFee"));
        oe->setProperty("EntryFee", oe->getDCI().getInt("EntryFee"));
        oe->setProperty("YouthFee", oe->getDCI().getInt("YouthFee"));

        oe->setProperty("Account", oe->getDCI().getString("Account"));
        oe->setProperty("LateEntryFactor", oe->getDCI().getString("LateEntryFactor"));
      }
      loadPage(gdi);
    }
    else if (bi.id == "Exit") {
      PostMessage(gdi.getMain(), WM_CLOSE, 0, 0);
    }
    else if (bi.id=="Browse") {
      string f = gdi.browseForOpen("XML-filer*.xml", "xml");
      string id;
      InputInfo *ii;
      if (!f.empty()) {
        ii = static_cast<InputInfo *>(bi.getExtra());
        if (ii)
          gdi.setText(ii->id, f);
      }
    }
    else if (bi.id=="DBaseIn") {
      gdi.clearPage(true);
      gdi.addString("", boldText, "Importera lpare och klubbar / distriktsregister");
      gdi.addString("", 10, "help:687188");
      gdi.dropLine(2);
      gdi.pushX();
      gdi.fillRight();
      InputInfo *ii = &gdi.addInput("ClubFile", "", 20, 0, "Filnamn IOF-XML med klubbar");
      gdi.dropLine();
      gdi.addButton("Browse", "Blddra...", CompetitionCB).setExtra(ii);
      gdi.popX();
      gdi.dropLine(3);
      ii = &gdi.addInput("CmpFile", "", 20, 0, "Filnamn IOF-XML med lpare");
      gdi.dropLine();
      gdi.addButton("Browse", "Blddra...", CompetitionCB).setExtra(ii);
      
      gdi.dropLine(4);

      gdi.popX();        
      gdi.addCheckbox("Clear", "Nollstll databaser", 0, true);
      gdi.dropLine(2);

      gdi.popX();
      gdi.addButton("DoDBaseIn", "Importera", CompetitionCB);
      gdi.addButton("Cancel", "Avbryt", CompetitionCB);
      gdi.dropLine(3);
      gdi.fillDown();
      gdi.popX();
    }
    else if (bi.id=="DoDBaseIn") {
      gdi.enableEditControls(false);
      gdi.disableInput("DoDBaseIn");
      gdi.disableInput("Cancel");

      gdi.setWaitCursor(true);
      gdi.addString("", 0, "Importerar...");
      bool clear = gdi.isChecked("Clear");
      oe->importXML_IOF_Data(gdi.getText("ClubFile").c_str(),
                             gdi.getText("CmpFile").c_str(), clear);
      
      gdi.dropLine();
      gdi.addButton("Cancel", "terg", CompetitionCB);
      gdi.setWaitCursor(false);                       
    }
    else if (bi.id=="Reset") {
      if (gdi.ask("Vill d terstlla instllningar och skriva ver egna databaser?"))
        Setup(true);
    }
		else if (bi.id=="ConnectMySQL") 
      loadConnectionPage(gdi);
    else if(bi.id=="SaveClient") { 
      oe->setClientName(gdi.getText("ClientName"));
      if (gdi.getText("ClientName").length()>0) 
          oe->setProperty("Client", gdi.getText("ClientName"));
    }
    else if (bi.id=="ConnectToMySQL") {
			bool s=oe->connectToMySQL(gdi.getText("Server"), 
                                gdi.getText("UserName"),
                                gdi.getText("PassWord"), 
                                gdi.getTextNo("Port"));

      if (s) {
        defaultServer=gdi.getText("Server");
        defaultName=gdi.getText("UserName");
        defaultPwd=gdi.getText("PassWord");
        defaultPort=gdi.getText("Port");

        oe->setClientName(gdi.getText("ClientName"));
        oe->setProperty("Server", defaultServer);
        oe->setProperty("UserName", defaultName);
        oe->setProperty("Port", defaultPort);
        if (gdi.getText("ClientName").length()>0) 
          oe->setProperty("Client", gdi.getText("ClientName"));


        loadConnectionPage(gdi);
      }
		}
    else if (bi.id=="DisconnectMySQL") {
      oe->closeDBConnection();
      loadConnectionPage(gdi);
		}
    else if (bi.id=="UploadCmp") {
      if(oe->uploadSynchronize())
        gdi.setWindowTitle(oe->getTitleName());

			loadConnectionPage(gdi);
    }
		else if(bi.id=="Cancel"){
			loadPage(gdi);
		}
    else if(bi.id=="dbtest") {
      
    }
    else if(bi.id=="FreeImport") {
      gdi.clearPage(true);
      gdi.addString("", 2, "Fri anmlningsimport");
      gdi.addString("", 10, "help:33940");
      gdi.dropLine(0.5);
      gdi.addInputBox("EntryText", 450, 280, entryText, 0, "");
      gdi.dropLine(0.5);
      gdi.fillRight();
      gdi.addButton("PreviewImport", "Granska inmatning", CompetitionCB, "tooltip:analyze");
      gdi.addButton("Paste", "Klistra in", CompetitionCB, "tooltip:paste");
      gdi.addButton("ImportFile", "Importera fil...", CompetitionCB, "tooltip:import");
      gdi.addButton("ImportDB", "Bygg databaser...", CompetitionCB, "tooltip:builddata");
      gdi.fillDown();
    }
    else if(bi.id=="ImportDB") {
      if(!gdi.ask("help:146122"))
        return 0;

      string file=gdi.browseForOpen("XML-filer*.xml;*.meos", "xml");

      if(file.empty())
        return 0;
      gdi.setWaitCursor(true);
      oe->getFreeImporter(fi);
      string info;

      oe->importXMLNames(file.c_str(), fi, info);
      fi.save();
      gdi.alert(info);
      gdi.setWaitCursor(false);
    }
    else if(bi.id=="Paste") {
      gdi.pasteText("EntryText");
    }
    else if(bi.id=="ImportFile") {
      string file=gdi.browseForOpen("Textfiler*.txt", "txt");
      ifstream fin(file.c_str());
      char bf[1024];
      bf[0]='\r';
      bf[1]='\n';
      entryText.clear();
      while (fin.good() && !fin.eof()) {
        fin.getline(bf+2, 1024-2);
        entryText+=bf;
      }
      entryText+="\r\n";
      fin.close();

      gdi.setText("EntryText", entryText);
    }
    else if(bi.id=="PreviewImport") {
      oe->getFreeImporter(fi);
      entryText=gdi.getText("EntryText");
      gdi.clearPage(false);
      gdi.addString("", 2, "Frhandsgranskning, import");
      gdi.dropLine(0.5);
      char *bf=new char[entryText.length()+1];
      strcpy_s(bf, entryText.length()+1, entryText.c_str());      
      fi.extractEntries(bf, entries);
      delete[] bf;
      fi.showEntries(gdi, entries);
      gdi.fillRight();
      gdi.dropLine(1);
      gdi.addButton("DoFreeImport", "Spara anmlningar", CompetitionCB);
      gdi.addButton("FreeImport", "ndra", CompetitionCB);
      
      gdi.addButton("Cancel", "Avbryt", CompetitionCB);
      gdi.scrollToBottom();
    }
    else if (bi.id=="DoFreeImport") {
      fi.addEntries(oe, entries);
      entryText.clear();
      loadPage(gdi);
    }
		else if(bi.id=="Startlist") {
      save(gdi, false);
			int FilterIndex = 1;
			string save=browseForSave("IOF Startlista (xml)*.xmlOE Semikolonseparerad (csv)*.csvWebben (html)*.html", "xml", FilterIndex);

			if(save.length()>0)	{
				gdi.setWaitCursor(true);
        oe->sanityCheck(gdi, false);
        if (FilterIndex == 1) {
				  oe->exportIOFStartlist(save.c_str());
        }
        else if (FilterIndex == 2) {
				  oe->exportOECSV(save.c_str());
        }
        else {
          oListParam par;
          par.listCode = EStdStartList;
          par.legNumber = -1;
          oListInfo li;
          oe->generateListInfo(par,  gdi.getLineHeight(), li);
          gdioutput tGdi(gdi.getScale());
          oe->generateList(tGdi, true, li);
          tGdi.writeTableHTML(save, oe->getName());
          tGdi.openDoc(save.c_str());
        }
          //gdi.alert("Not implemented");
			}
		}
		else if(bi.id=="Splits") {
      save(gdi, false);
			int FilterIndex=1;
			string save=browseForSave("IOF Resultat (XML)*.xmlOE Semikolonseparerad (csv)*.csvWebben (html)*.html", "xml", FilterIndex);

			if(save.length()>0)	{
        oe->sanityCheck(gdi, true);
				gdi.setWaitCursor(true);
        if (FilterIndex == 1) 
  				oe->exportIOFSplits(save.c_str());
        else if (FilterIndex == 2) {
				  oe->exportOECSV(save.c_str());
        }
        else
          gdi.alert("Not implemented");

			}
		}
		else if (bi.id=="ExportCSV") {
      save(gdi, false);
			int FilterIndex=1;
			string save=browseForSave("OE2003 (CSV)*.csv", "csv", FilterIndex);

			if(save.length()>0)	{
        oe->sanityCheck(gdi, false);
				gdi.setWaitCursor(true);
				oe->exportOECSV(save.c_str());
			}
		}
		else if (bi.id=="SaveAs") {
      oe->sanityCheck(gdi, false);
      save(gdi, true);
			exportFileAs(hWndMain, gdi);
		}
		else if (bi.id=="Import") {
      //Import complete competition
			importFile(hWndMain, gdi);     
			loadPage(gdi);	
		}
    else if (bi.id=="Restore") {
      char bf[260];
		  getUserFile(bf, "");
      gdi.clearPage(false);
      oe->enumerateBackups(bf);

      gdi.addString("", boldLarge, "Lagrade skerhetskopior");
      gdi.addString("", 10, "help:restore_backup");
      gdi.addButton("Cancel", "Avbryt", CompetitionCB);
      
      oe->listBackups(gdi, ::restoreCB);
      gdi.refresh();
    }
    else if(bi.id=="Save") {
      save(gdi, true);
      resetSaveTimer();
    }
		else if (bi.id=="CloseCmp") {
			gdi.setWaitCursor(true);
			if(!showConnectionPage)
        save(gdi, false);
      oe->save();
			oe->newCompetition("");
      resetSaveTimer();
      gdi.setWindowTitle("");
			if(showConnectionPage)
        loadConnectionPage(gdi);
      else
        loadPage(gdi);
			gdi.setWaitCursor(false);
		}
		else if (bi.id=="Delete" && 
			gdi.ask("Vill du verkligen radera tvlingen?")) {
			
      if (oe->isClient())
        oe->dropDatabase();
			else if (!oe->deleteCompetition()) 
				gdi.alert("Operation failed. It is not possible to delete competitions on server");
			
      oe->clearListedCmp();
			oe->newCompetition("");
      gdi.setWindowTitle("");
			loadPage(gdi);	
		}
		else if(bi.id=="NewCmp") {
			oe->newCompetition(lang.tl("Ny tvling"));
			gdi.setWindowTitle("");
      loadPage(gdi);
			return 0;			
		}
		else if (bi.id=="OpenCmp") {
			ListBoxInfo lbi;
      int id=0;
      bool frontPage=true;
			if (!gdi.getSelectedItem("CmpSel", &lbi)) {
        frontPage=false;
        if( gdi.getSelectedItem("ServerCmp", &lbi) )
          id=lbi.data;
        else if( gdi.getSelectedItem("LocalCmp", &lbi) )
          id=lbi.data;
      }
      else id=lbi.data;

      if(id==0)
        throw std::exception("Ingen tvling vald");

			gdi.setWaitCursor(true);
			if(!oe->open(id))
				gdi.alert("Kunde inte ppna tvlingen.");
			else {
				if(gSI) gSI->SetZeroTime(oe->getZeroTimeNum());
        resetSaveTimer();
        oe->setProperty("LastCompetition", id);
        gdi.setWindowTitle(oe->getTitleName());
        oe->updateTabs();

        if(frontPage)
				  loadPage(gdi);
        else {
          oe->verifyConnection();
          oe->validateClients();
          loadConnectionPage(gdi);
        }
        return 0;
			}
		}
		else if(bi.id=="BrowseCourse") {
			string file=browseForOpen("OCAD, banexport*.csv;*.txt;*.xml", "csv");
			if(file.length()>0)
				gdi.setText("FileName", file);
		}
		else if (bi.id=="BrowseEntries") {
			string file=browseForOpen("Importerbara*.xml;*.csvIOF-xml*.xmlSemikolonseparerad OE-standard*.csv", "xml");
      if(file.length()>0) {
        const char *ctrl = (const char *)bi.getExtra();
				if (ctrl != 0)
          gdi.setText(ctrl, file);
      }
		}
		else if(bi.id=="Entries") {
      if(!save(gdi, false))
        return 0;

			gdi.clearPage(true);
			gdi.addString("", 2, "Importera tvlingsdata");
			gdi.addString("", 10, "help:import_entry_data");
			gdi.dropLine();

      gdi.pushX();
			
			gdi.fillRight();
			gdi.addInput("FileNameCmp", "", 32, 0, "Tvlingsinstllningar (IOF-XML)");
			gdi.dropLine();
      gdi.addButton("BrowseEntries", "Blddra...", CompetitionCB).setExtra("FileNameCmp");
			gdi.popX();

      gdi.dropLine(2.5);
      gdi.addInput("FileNameCls", "", 32, 0, "Klasser (IOF-XML)");
			gdi.dropLine();
      gdi.addButton("BrowseEntries", "Blddra...", CompetitionCB).setExtra("FileNameCls");
			gdi.popX();

      gdi.dropLine(2.5);
      gdi.addInput("FileName", "", 32, 0, "Anmlningar (IOF-XML eller OE-CSV)");
			gdi.dropLine();
      gdi.addButton("BrowseEntries", "Blddra...", CompetitionCB).setExtra("FileName");
			gdi.popX();

      gdi.dropLine(2.5);
      gdi.addInput("FileNameRank", "", 32, 0, "Ranking (IOF-XML)");
			gdi.dropLine();
      gdi.addButton("BrowseEntries", "Blddra...", CompetitionCB).setExtra("FileNameRank");
			gdi.popX();

      gdi.dropLine(3);
			gdi.pushX();
			gdi.fillRight();
			gdi.addButton("DoImport", "Importera", CompetitionCB);
			gdi.fillDown();			
			gdi.addButton("Cancel", "Avbryt", CompetitionCB);
			gdi.popX();
		}
		else if(bi.id=="DoImport") {
			csvparser csv;
      gdi.enableEditControls(false);
			gdi.disableInput("DoImport");
			gdi.disableInput("Cancel");
			gdi.disableInput("BrowseEntries");

      string filename[4];
      filename[0] = gdi.getText("FileNameCmp");
      filename[1] = gdi.getText("FileNameCls");
      filename[2] = gdi.getText("FileName");
      filename[3] = gdi.getText("FileNameRank");

      for (int i = 0; i<4; i++) {
        if (filename[i].empty())
          continue;

        gdi.addString("", 0, "Behandlar: X#" + filename[i]);

        int type=csv.iscsv(filename[i].c_str());
  			
        if (type) {
          const char *File = filename[i].c_str();

				  if (type==1) {
					  gdi.addString("", 0, "Importerar OE2003 csv-fil...");
					  gdi.refresh();
					  gdi.setWaitCursor(true);
					  if (csv.ImportOE_CSV(*oe, File))	{
              gdi.addString("", 0, "Klart. X deltagare importerade.#" + itos(csv.nimport));
					  }
					  else gdi.addString("", 0, "Frsket misslyckades.");
				  }
          else if (type==2) {
					  gdi.addString("", 0, "Importerar OS2003 csv-fil...");
					  gdi.refresh();
					  gdi.setWaitCursor(true);
					  if (csv.ImportOS_CSV(*oe, File)) {						  
              gdi.addString("", 0, "Klart. X lag importerade.#" + itos(csv.nimport));
					  }
					  else gdi.addString("", 0, "Frsket misslyckades.");
          }
          else if (type==3) {
					  gdi.addString("", 0, "Importerar RAID patrull csv-fil...");
					  gdi.setWaitCursor(true);
					  if (csv.ImportRAID(*oe, File)) {						  
              gdi.addString("", 0, "Klart. X patruller importerade.#" + itos(csv.nimport));
					  }
					  else gdi.addString("", 0, "Frsket misslyckades.");

          }
			  }
			  else {
          oe->importXML_EntryData(gdi, filename[i].c_str());
			  }
        gdi.setWindowTitle(oe->getTitleName());
        oe->updateTabs();
      }
		  gdi.refresh();
		  gdi.addButton("Cancel", "OK", CompetitionCB);				
		}
		else if (bi.id=="Courses") {
      if(!save(gdi, false))
        return 0;
			gdi.clearPage(true);
			gdi.addString("", 2, "Importera banor/klasser");
			gdi.addString("", 0, "help:importcourse");
			gdi.dropLine();

			gdi.fillRight();
			gdi.pushX();
			gdi.addInput("FileName", "", 32, 0, "Filnamn");
			gdi.dropLine();
			gdi.fillDown();
			gdi.addButton("BrowseCourse", "Blddra...", CompetitionCB);
			
			gdi.popX();
			gdi.pushX();
			gdi.dropLine();
			gdi.fillRight();
			gdi.addButton("DoImportCourse", "Importera", CompetitionCB);
			gdi.fillDown();					
			gdi.addButton("Cancel", "Avbryt", CompetitionCB);
			gdi.popX();
		}
		else if (bi.id=="DoImportCourse") {
			gdi.disableInput("DoImportCourses");
			gdi.disableInput("Cancel");
			gdi.disableInput("BrowseCourses");
      string filename = gdi.getText("FileName");
      runCourseImport(gdi, filename, oe);
			gdi.addButton("Cancel", "OK", CompetitionCB);			
		}
    else if (bi.id=="About") {
      loadAboutPage(gdi);
    }
	}
  else if(type==GUI_LISTBOX) {
    ListBoxInfo lbi=*(ListBoxInfo *)data;

    if(lbi.id=="LocalCmp") 
      gdi.selectItemByData("ServerCmp", -1);
    else if(lbi.id=="ServerCmp") 
      gdi.selectItemByData("LocalCmp", -1);
    else if (lbi.id=="TextSize") {
      int textSize = lbi.data;
      oe->setProperty("TextSize", textSize);
      gdi.setFont(textSize, oe->getPropertyString("TextFont", "Arial"));
      loadPage(gdi);
    }
    else if (lbi.id == "Language") {
      lang.impl->loadTable(lbi.data, lbi.text);
      oe->updateTabs(true);
      oe->setProperty("Language", lbi.data);
      loadPage(gdi);
    }
  }
  else if (type==GUI_EVENT) {
    if( ((EventInfo*)data)->id=="Connections" ) {
      string s=gdi.getText("ClientName");
      loadConnectionPage(gdi);
      gdi.setText("ClientName", s);

    }
  }
  else if(type==GUI_CLEAR) {
    if (gdi.isInputChanged("")) {
      string name=gdi.getText("Name");

      if(!name.empty() && !oe->empty())
        save(gdi, false);
    }
    return 1;
	}
  return 0;
}


int TabCompetition::restoreCB(gdioutput &gdi, int type, void *data) {
  BackupInfo *bi =(BackupInfo *)((TextInfo *)(data))->getExtra();
  string fi(bi->FullPath);
  if (!oe->open(fi.c_str(), false)) {
    gdi.alert("Kunde inte ppna tvlingen.");
  }
  else {
		if(gSI) gSI->SetZeroTime(oe->getZeroTimeNum());

    const string &name = oe->getName();
    if (name.find_last_of("}") != name.length()-1)
      oe->setName(name + " {" + lang.tl("terstlld") +"}");
    
    oe->restoreBackup();

    gdi.setWindowTitle(oe->getTitleName());
    oe->updateTabs();
    resetSaveTimer();
		loadPage(gdi);
  }
  return 0;
}

void TabCompetition::copyrightLine(gdioutput &gdi) const
{   
  gdi.pushX();
  gdi.fillRight();
 
  gdi.addButton("About", "Om MeOS...", CompetitionCB);

  gdi.dropLine(0.4);
  gdi.fillDown(); 
  gdi.addString("", 0, MakeDash("#Copyright  2007-2010 Melin Software HB"));
  char bf[256];
  gdi.dropLine(1);
  gdi.popX();

  sprintf_s(bf, "#Version 2.5.%d, BETA %s", getMeosBuild(), getMeosDate().c_str());
  gdi.addString("", 0, bf).setColor(colorDarkRed);
  gdi.dropLine(0.2);
}

void TabCompetition::loadAboutPage(gdioutput &gdi) const
{
  gdi.clearPage(false);
  gdi.addString("", 2, MakeDash("Om MeOS - ett Mycket Enkelt OrienteringsSystem")).setColor(colorDarkBlue);
  gdi.dropLine(2);
  gdi.addStringUT(1, MakeDash("Copyright  2007-2010 Melin Software HB"));
  gdi.dropLine();
  gdi.addStringUT(10, "The database connection used is MySQL++\nCopyright "
                        "(c) 1998 by Kevin Atkinson, (c) 1999, 2000 and 2001 by MySQL AB,"
                        "\nand (c) 2004-2007 by Educational Technology Resources, Inc.\n"
                        "The database used is MySQL, Copyright (c) 2008 Sun Microsystems, Inc.");
  
  gdi.dropLine();
  gdi.addString("", 0, "Det hr programmet levereras utan ngon som helst garanti. Programmet r ");
  gdi.addString("", 0, "fritt att anvnda och du r vlkommen att distribuera det under vissa villkor,");
  gdi.addString("", 0, "se license.txt som levereras med programmet.");
	
  gdi.dropLine();
  gdi.addString("", 1, "Vi stder MeOS");
  vector<string> supp;
  getSupporters(supp);
  for (size_t k = 0; k<supp.size(); k++) 
    gdi.addStringUT(0, supp[k]);

  gdi.dropLine();
  gdi.addButton("Cancel", "Stng", CompetitionCB);
  gdi.refresh();
}

bool TabCompetition::loadPage(gdioutput &gdi)
{
  showConnectionPage=false;
  oe->checkDB();
	gdi.clearPage(true);
	gdi.fillDown();

	if (oe->empty()) {
		gdi.addString("", 2, "Vlkommen till MeOS");
    gdi.addString("", 1, MakeDash("#- ")+ lang.tl("ett Mycket Enkelt OrienteringsSystem")).setColor(colorDarkBlue);
		gdi.dropLine();

    gdi.fillRight();
    gdi.pushX();

    gdi.addSelection("CmpSel", 300, 400, CompetitionCB, "Vlj tvling:");

		char bf[260];
		getUserFile(bf, "");
		oe->enumerateCompetitions(bf, "*.meos");
		oe->fillCompetitions(gdi, "CmpSel",0);
		
    gdi.selectFirstItem("CmpSel");
    
    int lastCmp = oe->getPropertyInt("LastCompetition", 0);
    gdi.selectItemByData("CmpSel", lastCmp);

    gdi.dropLine(0.8);
    gdi.addButton("OpenCmp", "ppna", CompetitionCB).setDefault();
		
    gdi.dropLine(4);
		gdi.popX();
    
		gdi.addButton("NewCmp", "Ny tvling", CompetitionCB);
		gdi.addButton("Import", "Importera tvling...", CompetitionCB);
		gdi.addButton("Restore", "terstll skerhetskopia...", CompetitionCB);
		gdi.addButton("ConnectMySQL", "Databasanslutning...", CompetitionCB);

    gdi.popX();
    gdi.dropLine(3);
    
    gdi.dropLine(0.3);
    textSizeControl(gdi); 
    
    gdi.setCX(gdi.getCX()+gdi.getLineHeight()*2);
    gdi.dropLine(0.6);
    gdi.addButton("Setup", "Instllningar...", CompetitionCB);
    
    gdi.popX();
    gdi.fillDown();	
    gdi.dropLine(5);
    copyrightLine(gdi);

    gdi.addButton(gdi.GetPageX()-gdi.scaleLength(180), 
                  gdi.getCY()-gdi.getButtonHeight(), 
                  "Exit", "Avsluta", CompetitionCB);
    gdi.setInputFocus("CmpSel", true);
	}
	else {
		gdi.addString("", 3, "MeOS");
		gdi.dropLine();

    oe->synchronize();

		gdi.addInput("Name", oe->getName(), 24, 0, "Tvlingsnamn:");

		gdi.pushX();
		gdi.fillRight();
		gdi.addInput("Date", oe->getDate(), 8, 0, "Datum:");	
		gdi.fillDown();		
		gdi.addInput("ZeroTime", oe->getZeroTime(), 8, 0, "Nolltid:");

    if (oe->isClient()) 
      gdi.disableInput("ZeroTime");

		gdi.popX();
		gdi.addString("", 0, "help:zero_time");

    gdi.fillRight();
    gdi.dropLine();
    gdi.addButton("Settings", "Tvlingsinstllningar", CompetitionCB);
		gdi.addButton("Report", "Tvlingsrapport", CompetitionCB);
	  
    gdi.fillDown();
    gdi.popX();

    gdi.dropLine(3);

    gdi.fillDown();
    gdi.addCheckbox("UseEconomy", "Hantera klubbar och ekonomi", CompetitionCB, oe->useEconomy());
    
    gdi.addCheckbox("UseSpeaker", "Anvnd speakerstd", CompetitionCB, oe->getDCI().getInt("UseSpeaker")!=0);

    gdi.dropLine();
    textSizeControl(gdi);
    
    gdi.dropLine(4);
    gdi.popX();
    gdi.fillRight();
		gdi.addButton("Save", "Spara", CompetitionCB, "help:save");
    gdi.addButton("Delete", "Radera", CompetitionCB);
		gdi.addButton("CloseCmp", "Stng", CompetitionCB);
		
		gdi.dropLine(3);
    gdi.popX();

		gdi.fillRight(); 
    gdi.addButton("SaveAs", "Skerhetskopiera", CompetitionCB);
		gdi.addButton("ConnectMySQL", "Databasanslutning", CompetitionCB);

    gdi.dropLine(3);
    gdi.popX();


    gdi.fillDown();
    gdi.popX();

		gdi.newColumn();
    gdi.dropLine(5);
    gdi.setCX(gdi.getCX()+gdi.scaleLength(60));

    RECT rc;
    rc.top = gdi.getCY() - gdi.scaleLength(30);
    rc.left = gdi.getCX() - gdi.scaleLength(30);

    int bw = gdi.scaleLength(150);
    gdi.addString("", 1, "Importera tvlingsdata");
    gdi.addButton(gdi.getCX(), gdi.getCY(), bw, "Entries", "Anmlningar", 
                  CompetitionCB, "",  false, false);
		gdi.addButton(gdi.getCX(), gdi.getCY(), bw, "FreeImport", "Fri anmlningsimport", 
                  CompetitionCB, "", false, false);
    gdi.addButton(gdi.getCX(), gdi.getCY(), bw, "Courses", "Banor",
                  CompetitionCB, "", false, false);
		
    gdi.dropLine();
		gdi.addString("", 1, "Exportera tvlingsdata");
    gdi.addButton(gdi.getCX(), gdi.getCY(), bw, "Startlist", "Startlista", 
                  CompetitionCB, "Exportera startlista p fil", false, false);
		gdi.addButton(gdi.getCX(), gdi.getCY(), bw, "Splits", "Resultat && strcktider", 
                  CompetitionCB, "Exportera resultat p fil", false, false);
		
		gdi.dropLine();
    gdi.addString("", 1, "Lpardatabasen");
		gdi.addButton(gdi.getCX(), gdi.getCY(), bw, "DBaseIn", "Importera", 
                  CompetitionCB, "", false, false);
		gdi.addButton(gdi.getCX(), gdi.getCY(), bw, "ExportSetup", "Exportera",
                  CompetitionCB, "", false, false);

    rc.bottom = gdi.getCY() + gdi.scaleLength(30);
    rc.right = rc.left + bw + gdi.scaleLength(60);
  	
    gdi.addRectangle(rc, colorLightBlue);

    gdi.popX();
    gdi.dropLine(5);
    copyrightLine(gdi);

    gdi.setOnClearCb(CompetitionCB);
	}
	gdi.refresh();
  return true;
}

void TabCompetition::textSizeControl(gdioutput &gdi) const 
{
  int s = oe->getPropertyInt("TextSize", 0);
  const char *id="TextSize";
  gdi.fillRight();
  //gdi.addString("", 0, "Textstorlek:");
  gdi.dropLine(-0.5);
  gdi.addSelection(id, 90, 200, CompetitionCB, "Textstorlek:");
  gdi.addItem(id, lang.tl("Normal"), 0);
  gdi.addItem(id, lang.tl("Stor"), 1);
  gdi.addItem(id, lang.tl("Strre"), 2);
  gdi.addItem(id, lang.tl("Strst"), 3);
  gdi.selectItemByData(id, s);

  id = "Language";
  gdi.addSelection(id, 90, 200, CompetitionCB, "Sprk:");
  gdi.addItem(id, "Svenska", 103);
  gdi.addItem(id, "English", 104);

  int lang = oe->getPropertyInt("Language", 103);
  gdi.selectItemByData(id, lang);
}
