/************************************************************************
    MeOS - Orienteering Software
    Copyright (C) 2009-2010 Melin Software HB
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Melin Software HB - software@melin.nu - www.melin.nu
    Stigbergsvgen 7, SE-75242 UPPSALA, Sweden
    
************************************************************************/

#include "stdafx.h"
#include "Table.h"
#include "gdioutput.h"
#include "meos_util.h"
#include <exception>
#include <algorithm>
#include <set>
#include "oEvent.h"
#include "localizer.h"

extern HINSTANCE hInst;
const char *tId="_TABLE_SEL";

Table::Table(oEvent *oe_, int rowH, 
             const string &name, const string &iName)
{
  oe=oe_;
  tableName=name;
  internalName = iName;
  nTitles=0;
	PrevSort=-1;
  rowHeight=rowH;

  highRow=-1;
  highCol=-1;
  colSelected=-1;

  editRow=-1;
  editCol=-1;
  drawFilterLabel=false;
  currentSortColumn=-1;
  hEdit=0;

  hdcCompatible=0;
  hbmStored=0;
}

Table::~Table(void)
{
  if (hEdit)
    DestroyWindow(hEdit);
}

void Table::clearCellSelection() {
  upperRow = -1;
  lowerRow = -1;
  upperCol = -1;
  lowerCol = -1;
}

int Table::addColumn(const string &Title, int width, bool isnum)
{
	RowInfo ri;
	strcpy_s(ri.name, lang.tl(Title).c_str());
	ri.width=width;
	ri.isnumeric=isnum;
	Titles.push_back(ri);
  columns.push_back(nTitles);
	nTitles++;
  return Titles.size()-1;
}

void Table::moveColumn(int src, int target) 
{
  if(src==target)
    return;

  vector<int>::iterator it_s=find(columns.begin(), columns.end(), src);
  vector<int>::iterator it_t=find(columns.begin(), columns.end(), target);

  if(it_s!=columns.end()) {

    if(it_s<it_t) {
      columns.erase(it_s);
      it_t=find(columns.begin(), columns.end(), target);

      if(it_t!=columns.end())
        ++it_t;
      
      columns.insert(it_t, src);
    }
    else {
      columns.erase(it_s);
      it_t=find(columns.begin(), columns.end(), target);
      columns.insert(it_t, src);
    }
  }
}

void Table::addRow(int rowId)
{  
	TableRow tr(nTitles);
  tr.height=rowHeight;
  TableSortIndex tsi;

  if (Data.empty()) {
    sortIndex.clear();
    tsi.index=0;
    sortIndex.push_back(tsi);

    tsi.index=1;
    sortIndex.push_back(tsi);

    tsi.index=2;
    
    Data.resize(2, tr);

    for (unsigned i=0;i<nTitles;i++) {
      Data[0].cells[i].contents=Titles[i].name;
      Data[1].cells[i].contents="...";

      Data[0].cells[i].canEdit=false;
      Data[0].cells[i].type=cellEdit;
      Data[0].cells[i].owner=0;

      Data[1].cells[i].canEdit=false;
      Data[1].cells[i].type=cellEdit;
      Data[1].cells[i].owner=0;      
    }
  }
  else {
    tsi.index=sortIndex.size();
  }
  idToRow[rowId]=Data.size();
  sortIndex.push_back(tsi);
  Data.push_back(tr);
}

void Table::set(int column, oBase &owner, int id, const string &data, bool canEdit, CellType type)
{
  TableRow &row=Data.back();
    
  TableCell &cell=row.cells[column];
  
  cell.contents=data;
  cell.owner=&owner;
  cell.id=id;
  cell.canEdit=canEdit;
  cell.type=type;
}

void Table::filter(int col, const string &filt, bool forceFilter)
{
  const string &oldFilter=Titles[col].filter;
  vector<TableSortIndex> baseIndex;

  if (filt==oldFilter && (!forceFilter || filt.empty()))
    return;
  else if (strncmp(oldFilter.c_str(), filt.c_str(), oldFilter.length())==0) {
    //Filter more...
    baseIndex.resize(2);
    baseIndex[0]=sortIndex[0];
    baseIndex[1]=sortIndex[1];
    swap(baseIndex, sortIndex);
    Titles[col].filter=filt;
  }
  else {
    //Filter less -> refilter all!
    Titles[col].filter=filt;
    sortIndex.resize(Data.size());
    TableSortIndex tsi;
    for (size_t k=0; k<Data.size();k++) {      
      tsi.index=k;
      sortIndex[k]=tsi;
    }

    for (unsigned k=0;k<nTitles;k++) {
      filter(k, Titles[k].filter, true);
    }
    return;
  }  

  char filt_lc[1024];
  strcpy_s(filt_lc, filt.c_str());
  CharLowerBuff(filt_lc, filt.length());

  sortIndex.resize(2);
  for (size_t k=2;k<baseIndex.size();k++) {
    if (filterMatchString(Data[baseIndex[k].index].cells[col].contents, filt_lc))
      sortIndex.push_back(baseIndex[k]);
  }
}

void Table::sort(int col)
{
  if(sortIndex.size()<2)
    return;
  currentSortColumn=col;
	if (PrevSort!=col && PrevSort!=-(10+col)) {
	  if(Titles[col].isnumeric)
	  {
      char bf[16];
		  //Nummeric data
      for(size_t k=2; k<sortIndex.size(); k++){					
        sprintf_s(bf, "%08X", atoi(Data[sortIndex[k].index].cells[col].contents.c_str()));
        sortIndex[k].key=bf;
		  }
    }
    else {		
      for(size_t k=2; k<sortIndex.size(); k++){					      
        sortIndex[k].key=Data[sortIndex[k].index].cells[col].contents;
		  }
    }
    std::stable_sort(sortIndex.begin()+2, sortIndex.end());
		PrevSort=col;
	} 
	else {
    std::reverse(sortIndex.begin()+2, sortIndex.end());

    if(PrevSort==col)
		  PrevSort=-(10+col);
    else
      PrevSort=col;
	}
}

int TablesCB(gdioutput *gdi, int type, void *data)
{
	if(type!=GUI_LINK || gdi->Tables.empty())
		return 0;

	TableInfo &tai=gdi->Tables.front();
	Table *t=tai.table;

	TextInfo *ti=(TextInfo *)data;


	if(ti->id.substr(0,4)=="sort"){
		int col=atoi(ti->id.substr(4).c_str());
		t->sort(col);
	}

	gdi->refresh();
	//gdi->Restore();
	//t->Render(*gdi);
//*/
	return 0;
}

void Table::getDimension(gdioutput &gdi, int &dx, int &dy) const
{
 	dy=rowHeight*(2+Data.size());

	dx=1;
  for(size_t i=0;i<columns.size();i++)
		dx+=Titles[columns[i]].width+1;
}

void Table::redrawCell(gdioutput &gdi, HDC hDC, int c, int r) 
{
  if (unsigned(r)<Data.size() && unsigned(c)<unsigned(nTitles)) {
    const TableRow &row=Data[r];
    const TableCell &cell=row.cells[c];
    RECT rc=cell.absPos;
    draw(gdi, hDC, table_xp+gdi.OffsetX, table_yp+gdi.OffsetY, rc);
  }
}

void Table::highlightCell(HDC hDC, const TableCell &cell, DWORD color, int dx, int dy)
{
  SelectObject(hDC, GetStockObject(DC_BRUSH));
  SelectObject(hDC, GetStockObject(NULL_PEN));
  
  SetDCBrushColor(hDC, color);

  RECT rc=cell.absPos;
  rc.left+=dx;
  rc.right+=dx;
  rc.top+=dy;
  rc.bottom+=dy;

  Rectangle(hDC, rc.left+1, rc.top, 
                 rc.right+2, rc.bottom);

  SelectObject(hDC, GetStockObject(DEFAULT_GUI_FONT));
  SetBkMode(hDC, TRANSPARENT);

  rc.left+=4;
  rc.top+=1;
  DrawText(hDC, cell.contents.c_str(), -1, &rc, DT_LEFT|DT_NOPREFIX);
}

void Table::startMoveCell(HDC hDC, const TableCell &cell)
{
  RECT rc=cell.absPos;
  int cx=rc.right-rc.left+1;
  int cy=rc.bottom-rc.top;

  hdcCompatible = CreateCompatibleDC(hDC); 
  hbmStored = CreateCompatibleBitmap(hDC, cx, cy); 
 
  // Select the bitmaps into the compatible DC. 
  SelectObject(hdcCompatible, hbmStored);

  lastX=rc.left;
  lastY=rc.top;
  BitBlt(hdcCompatible, 0, 0, cx, cy, hDC, lastX, lastY, SRCCOPY);
}

void Table::restoreCell(HDC hDC, const TableCell &cell)
{
  const RECT &rc=cell.absPos;
  int cx=rc.right-rc.left+1;
  int cy=rc.bottom-rc.top;

  if(hdcCompatible) {  
    //Restore bitmap
    BitBlt(hDC, lastX, lastY, cx, cy, hdcCompatible, 0, 0, SRCCOPY);
  }
}

void Table::moveCell(HDC hDC, const TableCell &cell, int dx, int dy)
{
  RECT rc=cell.absPos;
  rc.left+=dx; rc.right+=dx;
  rc.top+=dy;  rc.bottom+=dy;
  int cx=rc.right-rc.left+1;
  int cy=rc.bottom-rc.top;

  if(hdcCompatible) {  
    lastX=rc.left;
    lastY=rc.top;
    //Take new
    BitBlt(hdcCompatible, 0, 0, cx, cy, hDC, lastX, lastY , SRCCOPY);
  }
  else {
    startMoveCell(hDC, cell);
    dx=0;
    dy=0;
  }
  highlightCell(hDC, cell, RGB(255, 0,0), dx, dy);
}

void Table::stopMoveCell(HDC hDC, const TableCell &cell, int dx, int dy)
{
  if(hdcCompatible) {
    restoreCell(hDC, cell);
/*    RECT rc=cell.absPos;
    rc.left+=dx; rc.right+=dx;
    rc.top+=dy;  rc.bottom+=dy;
    int cx=rc.right-rc.left+2;
    int cy=rc.bottom-rc.top+2;

    //Restore bitmap
    BitBlt(hDC, lastX, lastY, cx, cy, hdcCompatible, 0, 0, SRCCOPY);
*/
    DeleteDC(hdcCompatible);
    hdcCompatible=0;
  }

  if(hbmStored) {
    DeleteObject(hbmStored);
    hbmStored=0;
  }
}

bool Table::mouseMove(gdioutput &gdi, int x, int y)
{
  int row=getRow(y);
  int col=-1;

  if(row!=-1)
    col=getColumn(x);

  HWND hWnd=gdi.getTarget();

  if(colSelected!=-1) {
    TableCell &cell=Data[0].cells[colSelected];
    HDC hDC=GetDC(hWnd);

    restoreCell(hDC, cell);

    if (col!=highCol) {
      if(unsigned(highRow)<Data.size() && unsigned(highCol)<nTitles) 
        redrawCell(gdi, hDC, highCol, highRow);

      DWORD c=RGB(240, 200, 140);      
      if(unsigned(col)<nTitles)
        highlightCell(hDC, Data[0].cells[col], c, 0,0);
    }
    
    //highlightCell(hDC, cell, RGB(255,0,0), x-startX, y-startY);
    moveCell(hDC, cell,  x-startX, y-startY);
    ReleaseDC(hWnd, hDC);
    highRow=0;
    highCol=col;
    return false;
  }

  RECT rc;
  GetClientRect(hWnd, &rc);

  if(x<=rc.left || x>=rc.right || y<rc.top || y>rc.bottom)
    row=-1;

  if (row>=0 && col>=0) {
    //int index=sortIndex[row].index;
    TableRow &trow=Data[row];
    TableCell &cell=trow.cells[col];
   
    if(highRow!=row || highCol!=col) {

      HDC hDC=GetDC(hWnd); 

      if(unsigned(highRow)<Data.size() && unsigned(highCol)<Titles.size()) 
        redrawCell(gdi, hDC, highCol, highRow);
 
      DWORD c;
      if(cell.canEdit)
        c=RGB(240, 240, 150);
      else
        c=RGB(240, 200, 140);
      
      highlightCell(hDC, cell, c, 0,0);

      ReleaseDC(hWnd, hDC);
      SetCapture(hWnd);
      highCol=col;
      highRow=row;
      return true;
    }
    return true;
  }  

  if(unsigned(highRow)<Data.size() && unsigned(highCol)<Titles.size()) {
    ReleaseCapture();
    HDC hDC=GetDC(hWnd);
    redrawCell(gdi, hDC, highCol, highRow);
    ReleaseDC(hWnd, hDC);
    highRow=-1;
  }
  return false;
}

bool Table::mouseLeftUp(gdioutput &gdi, int x, int y)
{
  if(colSelected!=-1) {

    if(hdcCompatible) {
      TableCell &cell=Data[0].cells[colSelected];
      HWND hWnd=gdi.getTarget();
      HDC hDC=GetDC(hWnd);
      stopMoveCell(hDC, cell, x-startX, y-startY);
      ReleaseDC(hWnd, hDC);
      //return true;
    }

    if (highRow==0 && colSelected==highCol) {
      colSelected=-1;
      sort(highCol);
      gdi.refresh();
      mouseMove(gdi, x, y);
      return true;
    }
    else {
      moveColumn(colSelected, highCol);
      InvalidateRect(gdi.getTarget(), 0, false);
      colSelected=-1;
      return true;
    }
  }

  colSelected=-1;
  return false;
}

int tblSelectionCB(gdioutput *gdi, int type, void *data) 
{
  if (type == GUI_LISTBOX) {
    ListBoxInfo lbi = *static_cast<ListBoxInfo *>(data);
    TableCell &cell = *static_cast<TableCell *>(lbi.getExtra());
    string output=cell.contents;
    cell.owner->inputData(cell.id, lbi.text, lbi.data, output, false);
    cell.contents=output;
    //gdi->hideControl(tId);
  }
  return 0;
}

bool Table::character(gdioutput &gdi, int charCode, int repeat) {
  return false;
}


bool Table::destroyEditControl(gdioutput &gdi) {
  colSelected=-1;

  if (hEdit) {
    if (!enter(gdi))
      return false;
    
    if (hEdit) {
      DestroyWindow(hEdit);
      hEdit=0;
    }
  }
  gdi.removeControl(tId);

  if (drawFilterLabel) {
    drawFilterLabel=false;
    gdi.refresh();
  }

  return true;
}

bool Table::mouseLeftDown(gdioutput &gdi, int x, int y) {
  clearCellSelection();
  if (!destroyEditControl(gdi))
    return false;

  if (highRow==0) {
    colSelected=highCol;
    startX=x;
    startY=y;
    //sort(highCol);
    //gdi.refresh();
    //mouseMove(gdi, x, y);
  }
  else if (highRow==1) {
    //filter(highCol, "lots");
    RECT rc=Data[1].cells[0].absPos; 
    //rc.right=rc.left+tableWidth;
    editRow=highRow;
    editCol=highCol;

    hEdit=CreateWindowEx(0, "EDIT", Titles[highCol].filter.c_str(), 
        WS_TABSTOP|WS_VISIBLE | WS_CHILD | ES_AUTOHSCROLL|WS_BORDER,
        rc.left+105, rc.top, tableWidth-105, (rc.bottom-rc.top-1), gdi.getTarget(),
        0, hInst, 0); 
    drawFilterLabel=true;
    SendMessage(hEdit, EM_SETSEL, 0, -1); 
    SetFocus(hEdit);
    SendMessage(hEdit, WM_SETFONT, (WPARAM) GetStockObject(DEFAULT_GUI_FONT), 0);
    gdi.refresh();
  }
  else
    SetFocus(gdi.getHWND());

  return false;
}

bool Table::mouseLeftDblClick(gdioutput &gdi, int x, int y)
{
  clearCellSelection();
  if (!destroyEditControl(gdi))
    return false;

  if (unsigned(highRow)>=Data.size() || unsigned(highCol)>=Titles.size()) {
    return false;
  }

 
  if (highRow != 0 && highRow != 1) {
    if (editCell(gdi, highRow, highCol))
      return true;
  }
  return false;
}

bool Table::editCell(gdioutput &gdi, int row, int col) {
  TableCell &cell = Data[row].cells[col];
    
  if (!cell.canEdit) {
    MessageBeep(-1);
    return true;
  }
  ReleaseCapture();

  editRow = row;
  editCol = col;

  RECT &rc=cell.absPos;

  if (cell.type==cellAction) {
    gdi.makeEvent("CellAction", cell.id, cell.owner);
    return true;
  }
  if (cell.type==cellSelection) {
    gdi.addSelection(rc.left+gdi.OffsetX, rc.top+gdi.OffsetY, tId, 
            rc.right-rc.left+1, (rc.bottom-rc.top)*10, 
            tblSelectionCB).setExtra(&cell); 
    cell.owner->fillInput(cell.id, gdi, tId);
    return true;
  }
  else if (cell.type==cellEdit) {
    hEdit=CreateWindowEx(0, "EDIT", cell.contents.c_str(), 
      WS_TABSTOP|WS_VISIBLE | WS_CHILD | ES_AUTOHSCROLL|WS_BORDER,
      rc.left, rc.top, rc.right-rc.left+1, (rc.bottom-rc.top), gdi.getTarget(),
      0, hInst, 0);    
    SendMessage(hEdit, EM_SETSEL, 0, -1); 
    SetFocus(hEdit);
    SendMessage(hEdit, WM_SETFONT, (WPARAM) GetStockObject(DEFAULT_GUI_FONT), 0);
    return true;
  }  

  return false;
}

void Table::draw(gdioutput &gdi, HDC hDC, int dx, int dy, const RECT &screen) 
{
  if (Data.empty()) {
    addRow(0);
    Data.resize(2, TableRow(nTitles));
    sortIndex.resize(2);
  }
	gdi.getLineHeight();
  highRow=-1;
  highCol=-1;
  table_xp=dx-gdi.OffsetX;
  table_yp=dy-gdi.OffsetY;

  if(currentSortColumn==-1)
    sort(0);

	int wi, he;
	getDimension(gdi, wi, he);

  tableWidth=wi;
  tableHeight=he;

  //Find first and last row
  int yreltop=(screen.top-(dy-gdi.OffsetY)-rowHeight+1)/rowHeight;
  int yrelbottom=(screen.bottom-(dy-gdi.OffsetY)-1)/rowHeight;

  int firstRow=max(yreltop,0);
  int lastRow=min(yrelbottom, int(sortIndex.size()));

  int firstCol=0;
  int lastCol=columns.size();

  xpos.resize(columns.size()+1);

  xpos[0]=dx-gdi.OffsetX;
  for (size_t i=1;i<=columns.size();i++) 
    xpos[i]=xpos[i-1]+Titles[columns[i-1]].width+1;
 
  //Find first and last visible column
  while(firstCol<int(columns.size()) && xpos[firstCol+1]<=screen.left)
    firstCol++;

  while(lastCol>0 && xpos[lastCol-1]>screen.right)
    lastCol--;

  SelectObject(hDC, GetStockObject(DEFAULT_GUI_FONT));
  SetBkMode(hDC, TRANSPARENT); 
	SelectObject(hDC, GetStockObject(DC_BRUSH));
	SelectObject(hDC, GetStockObject(DC_PEN));

	SetDCPenColor(hDC, RGB(190,190,190));
  SetDCBrushColor(hDC, RGB(60,25,150));
  //Title
  
  if(firstRow==0 && lastRow>=0) {
    Rectangle(hDC, dx-gdi.OffsetX, dy-gdi.OffsetY, 
              dx+tableWidth-gdi.OffsetX, dy+rowHeight-gdi.OffsetY);  
    
    RECT rc;
    rc.left=dx+5-gdi.OffsetX;
    rc.right=rc.left+150;
    rc.top=dy+3-gdi.OffsetY;
    rc.bottom=rc.top+rowHeight;
    SetTextColor(hDC, RGB(255,255,255));
    DrawText(hDC, lang.tl(tableName).c_str(), -1, &rc, DT_LEFT|DT_NOPREFIX);
    char bf[256];

    sprintf_s(bf, lang.tl("Sortering: %s, antal rader: %d.").c_str(), 
              Titles[currentSortColumn].name, int(sortIndex.size())-2);
    rc.left=rc.right+10;
    rc.right=tableWidth-gdi.OffsetX;
    DrawText(hDC, bf, -1, &rc, DT_LEFT|DT_NOPREFIX);

    SetTextColor(hDC, RGB(0,0,0));

    if (drawFilterLabel) {
      SetDCBrushColor(hDC, RGB(100,200,100));

      Rectangle(hDC, dx-gdi.OffsetX, dy+2*rowHeight-gdi.OffsetY, 
              dx+tableWidth-gdi.OffsetX, dy+3*rowHeight-gdi.OffsetY);  

      rc.left=dx+5-gdi.OffsetX;
      rc.right=rc.left+100;
      rc.top=dy+3+2*rowHeight-gdi.OffsetY;
      rc.bottom=rc.top+rowHeight;
      char tbf[2];
      tbf[0]=Titles[editCol].name[0];
      tbf[1]=0;
      CharLower(tbf);
      string filter = lang.tl("Urval %c%s:")+" ";
      sprintf_s(bf, filter.c_str(), tbf[0], 
              Titles[editCol].name+1, int(sortIndex.size())-2);
      
      DrawText(hDC, bf, -1, &rc, DT_RIGHT|DT_NOPREFIX);
    }
  }
  SetDCBrushColor(hDC, RGB(200,200,200));
  
  int yp;
  yp=dy+rowHeight;

  if(firstRow<=2 && lastRow>=1) {
    
 	  for (int k=firstCol;k<lastCol;k++) {
      int xp=xpos[k];
      int index=columns[k];
      RECT rc;
      rc.left=xp;
      rc.right=rc.left+Titles[index].width;
      rc.top=yp+1-gdi.OffsetY;
      rc.bottom=rc.top+rowHeight;
      Rectangle(hDC, rc.left, rc.top-1, rc.right+2, rc.bottom);  
      Data[0].cells[index].absPos=rc;
      DrawText(hDC, Titles[index].name, -1, &rc, DT_CENTER|DT_NOPREFIX);
      Titles[index].title=rc;

      if (!drawFilterLabel) {
        rc.top+=rowHeight;
        rc.bottom+=rowHeight; 
        Titles[index].condition=rc;
        Data[1].cells[index].absPos=rc;
        if(Titles[index].filter.empty()) {
          Rectangle(hDC, rc.left, rc.top-1, rc.right+2, rc.bottom);
          rc.left+=3;
          DrawText(hDC, lang.tl("Urval...").c_str(), -1, &rc, DT_LEFT|DT_NOPREFIX);
        }
        else { 
          SetDCBrushColor(hDC, RGB(230,100,100));  
          Rectangle(hDC, rc.left, rc.top-1, rc.right+2, rc.bottom);
          rc.left+=3;
          DrawText(hDC, Titles[index].filter.c_str(), -1, &rc, DT_LEFT|DT_NOPREFIX);
          SetDCBrushColor(hDC, RGB(200,200,200));  
        }
      }
    }
  }

	TextInfo ti;
	SelectObject(hDC, GetStockObject(DC_BRUSH));

	SetDCPenColor(hDC, RGB(190,190, 190));

  for (size_t k1=max(2, firstRow); int(k1)<lastRow; k1++){
    int yp=dy+rowHeight*(k1+1);
    TableRow &tr=Data[sortIndex[k1].index];
		int xp=xpos[0];

		if(k1&1)
			SetDCBrushColor(hDC, RGB(230,230, 240));
		else
			SetDCBrushColor(hDC, RGB(230,230, 250));

		SelectObject(hDC, GetStockObject(NULL_PEN));
    Rectangle(hDC, max(xpos[firstCol], int(screen.left)), yp-gdi.OffsetY, 
              min(xpos[lastCol]+1, int(screen.right+2)), yp+rowHeight-gdi.OffsetY);
  	SelectObject(hDC, GetStockObject(DC_PEN));
    const int cy=yp+rowHeight-gdi.OffsetY-1;
    MoveToEx(hDC, xp, cy, 0);
    LineTo(hDC, xp+wi, cy);
    
		for(int k=firstCol;k<lastCol;k++){
      xp=xpos[k];
      int i=columns[k];
      if(!tr.cells[i].contents.empty()) {
        ti.xp=xp+4+gdi.OffsetX;
			  ti.yp=yp+1;
			  ti.format=0;
			  ti.xlimit=Titles[i].width;
        gdi.RenderString(ti, tr.cells[i].contents, hDC);
      }
      tr.cells[i].absPos.left=xp;
      tr.cells[i].absPos.right=xp+Titles[i].width;
      tr.cells[i].absPos.top=yp-gdi.OffsetY;
      tr.cells[i].absPos.bottom=yp-gdi.OffsetY+rowHeight;
		}			
	}

  const int miy=dy+rowHeight-gdi.OffsetY;
  const int may=dy+rowHeight*(sortIndex.size()+1)-gdi.OffsetY;

 	for (size_t i=firstCol;i<=columns.size();i++) {
		int xp=xpos[i];
		MoveToEx(hDC, xp, miy, 0);
    LineTo(hDC, xp, may);
	}
}

void Table::print(gdioutput &gdi, HDC hDC, int dx, int dy)
{
  vector<int> widths(columns.size());
  vector<bool> skip(columns.size(), true);
  int rh = 0;

  for (size_t j=0;j<columns.size();j++) {
    int i=columns[j];
    TextInfo ti;
    ti.format = boldSmall;
    ti.text =  Titles[i].name;
    gdi.calcStringSize(ti, hDC);
    widths[j] = max<int>(widths[j], ti.TextRect.right-ti.TextRect.left);
    rh = max<int>(rh, ti.TextRect.bottom-ti.TextRect.top);
  }
  const int extra = 10;

  for (size_t j=0;j<columns.size();j++) {
    for (size_t k=2; k<sortIndex.size(); k++) {
      TableRow &tr=Data[sortIndex[k].index];
      const string &ct = tr.cells[columns[j]].contents;
      if (!ct.empty()) {
        skip[j] = false;
        TextInfo ti;
        ti.format = fontSmall;
        ti.text = ct;
        gdi.calcStringSize(ti, hDC);
        widths[j] = max<int>(widths[j], 
          int(1.1*(ti.TextRect.right-ti.TextRect.left)+extra));
        rh = max<int>(rh, ti.TextRect.bottom-ti.TextRect.top);
      }
    }
	}

  rh = int(rh*1.3);
  vector<int> adj_xp(columns.size());
  int w = widths[0];
  for (size_t j=1;j<columns.size();j++) 
    if (!skip[j]) {
      adj_xp[j] = adj_xp[j-1] + w +5;
      w = widths[j];
    }
    else
      adj_xp[j] = adj_xp[j-1];

  for (size_t j=0;j<columns.size();j++) {
    int yp=dy+rh;
    int xp=dx+adj_xp[j];
    int i=columns[j];
    if (!skip[j])
      gdi.addString("", yp, xp, boldSmall, Titles[i].name);
  }

  for (size_t k=2; k<sortIndex.size(); k++) {
    TableRow &tr=Data[sortIndex[k].index];
    int yp=dy+rh*k;
    
    for (size_t j=0;j<columns.size();j++) {
      if (!skip[j]) {
        int xp=dx+adj_xp[j];
        int i=columns[j];
        
        if(!tr.cells[i].contents.empty()) 
          gdi.addString("", yp, xp, fontSmall, tr.cells[i].contents, widths[j]);
      }
    }
	}
}

bool Table::UpDown(gdioutput &gdi, int direction)
{
  return false;
}

bool Table::tabFocus(gdioutput &gdi, int direction)
{
  if(hEdit) {
    DestroyWindow(hEdit);
    drawFilterLabel=false;
    gdi.refresh();
    return true;
  }

  return false;
}

bool Table::enter(gdioutput &gdi)
{
  if (hEdit) {
    if (unsigned(editRow)<Data.size() && unsigned(editCol)<Titles.size()) {            
      char bf[1024];
      GetWindowText(hEdit, bf, 1024);

      if (editRow>=2) {
        TableCell &cell=Data[editRow].cells[editCol];
        string output;

        try {
          cell.owner->inputData(cell.id, bf, 0, output, false); 
          cell.contents=output;
          DestroyWindow(hEdit);
          hEdit=0;
          return true;
        }
        catch(const std::exception &ex) {         
          string msg(ex.what());
          gdi.alert(msg);
        }
      }
      else if(editRow==1) {//Filter
        filter(editCol, bf);
        gdi.refresh();
        return true;
      }
    }
  }
  return false;
}

void Table::escape(gdioutput &gdi)
{
  if(hEdit) {
    DestroyWindow(hEdit);
    hEdit = 0;
  }
  gdi.removeControl(tId);
  drawFilterLabel=false;
  gdi.refresh();
}

bool Table::inputChange(gdioutput &gdi, HWND hdt)
{
  if (hEdit==hdt) {
  
    if (drawFilterLabel) {
      char bf[256];
      GetWindowText(hEdit, bf, 256);
      filter(editCol, bf);
      gdi.refresh();
    }

    return true;
  }
  return false;
}

int Table::getColumn(int x) const
{
  if(x<xpos[0])
    return -1;
  for(size_t k=0;k<columns.size();k++) {
    if(x<xpos[k+1])
      return columns[k];
  }
  return -1;
}

int Table::getRow(int y) const
{
  int r=(y-table_yp)/rowHeight-1;

  if(r>=0 && r<int(sortIndex.size()))
    return sortIndex[r].index;
  else return -1;
}

void Table::reloadRow(int rowId)
{
  if (idToRow.count(rowId)==0)
    throw std::exception("Row not found.");

  size_t index=idToRow[rowId];

  if(index>=Data.size())
    throw std::exception("Index out of bounds.");

  TableRow &row=Data[index];
  row.reload();
}

void TableRow::reload() 
{
  for (size_t k=0;k<cells.size();k++)
    cells[k].reload();
}

void TableCell::reload()
{
  if(owner)
    owner->getCellData(id, contents);
}

vector<pair<string, bool>> Table::getColumns() const
{
  vector<pair<string, bool>> cols(Titles.size());

  // Get all columns
  for (size_t k=0; k<cols.size(); k++) 
    cols[k].first = Titles[k].name;
  
  // Mark shown columns as selected
  for (size_t k=0; k<columns.size(); k++) 
    cols[columns[k]].second = true; 

  return cols;
}

void Table::selectColumns(const std::set<int> &sel)
{
  vector<int> newcols;

  for (size_t k=0; k<columns.size(); k++) {
    if (sel.count(columns[k])==1) {
      newcols.push_back(columns[k]);
    }
  }

  std::set<int>::const_iterator it = sel.begin();

  while(it != sel.end()) {
    if (count(newcols.begin(), newcols.end(), *it)==0)
      newcols.push_back(*it);
    ++it;
  }

  swap(columns, newcols);
}

void Table::resetColumns()
{
  columns.resize(Titles.size());

  for (size_t k=0;k<columns.size();k++)
    columns[k] = k;

  for (size_t k=0;k<nTitles;k++) 
    filter(k, "", false);

}

void Table::update() 
{
  Data.clear();
  sortIndex.clear();

  oe->generateTableData(internalName, this);

  //Refilter all
  for (size_t k=0;k<nTitles;k++) 
    filter(k, Titles[k].filter, true);
}
