#pragma once
/************************************************************************
    MeOS - Orienteering Software
    Copyright (C) 2009-2010 Melin Software HB
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Melin Software HB - software@melin.nu - www.melin.nu
    Stigbergsvgen 7, SE-75242 UPPSALA, Sweden
    
************************************************************************/

#include <vector>
#include <map>
#include <set>
#include "oBase.h"

enum CellType {cellEdit, cellSelection, cellAction};

class TableCell
{
  string contents;
  RECT absPos;
  DWORD id;
  oBase *owner;
  bool canEdit;
  CellType type;

  friend class TableRow;
  friend class Table;
  friend int tblSelectionCB(gdioutput *gdi, int type, void *data); 

  void reload();
};

class TableRow
{
protected:
	vector<TableCell> cells;

	string *SortString;
	int sInt;

  int ypos;
  int height;

  oBase *ob;
public:

  void reload();

	bool operator<(const TableRow &r){return *SortString<*r.SortString;}
	static bool cmpint(const TableRow &r1, const TableRow &r2) {return r1.sInt<r2.sInt;}

	TableRow(int elem): sInt(0)
	{
		cells.resize(elem);
		SortString=&cells[0].contents;
	}

	TableRow(const TableRow &t)
	{
		cells=t.cells;
    SortString=&cells[0].contents;
	}
	friend class Table;
};

class gdioutput;

struct RowInfo
{	
	char name[32];
	int width;
	bool isnumeric;
  RECT title;
  RECT condition;
  string filter;
};

struct TableSortIndex
{
  int index;
  //int ypos;
  string key;
  bool operator<(const TableSortIndex &t) const {return key<t.key;}
  bool operator<=(const TableSortIndex &t) const {return key<=t.key;}
};

class Table
{
protected:
  string tableName;
  string internalName;
	vector<RowInfo> Titles;
  vector<int> xpos;
	unsigned nTitles;
	int PrevSort;
	int rowHeight;
	vector<TableRow> Data;
  vector<TableSortIndex> sortIndex;
  vector<int> columns;
  map<int, int> idToRow;
  //highlight
  int highRow;
  int highCol;

  // Selected columns. For drag/drop and sort
  int colSelected;
  int startX;
  int startY;

  //Edit selection
  int editRow;
  int editCol;

  // Selected rectangle
  int upperRow;
  int lowerRow;
  int upperCol;
  int lowerCol;

  HWND hEdit;

  //For moving objects
  HDC hdcCompatible;
  HBITMAP hbmStored;
  int lastX;
  int lastY;

  //static bool filterMatchString(const string &c, const char *filt);
  void highlightCell(HDC hDC, const TableCell &cell, DWORD color, int dx, int dy);

  void moveCell(HDC hDC, const TableCell &cell, int dx, int dy);
  void startMoveCell(HDC hDC, const TableCell &cell);
  void stopMoveCell(HDC hDC, const TableCell &cell, int dx, int dy);
  void restoreCell(HDC hDC, const TableCell &cell);

  void moveColumn(int src, int target);

  int tableWidth;
  int tableHeight;
  bool drawFilterLabel;
  int currentSortColumn;

  int getColumn(int x) const;
  int getRow(int y) const;

  void redrawCell(gdioutput &gdi, HDC hDC, int c, int r);

  //Draw coordinates
  int table_xp;
  int table_yp;

  oEvent *oe;

  void clearCellSelection();


  bool destroyEditControl(gdioutput &gdi);

public:

  bool hasEditControl() {return hEdit!=0;}
  vector< pair<string, bool> > getColumns() const;
  void selectColumns(const set<int> &sel);

  oEvent *getEvent() const {return oe;}
	void getDimension(gdioutput &gdi, int &dx, int &dy) const;
	void draw(gdioutput &gdi, HDC hDC, int dx, int dy, 
            const RECT &screen);

  void print(gdioutput &gdi, HDC hDC, int dx, int dy);
  
  //Returns true if capture is taken
  bool mouseMove(gdioutput &gdi, int x, int y);
  bool mouseLeftDown(gdioutput &gdi, int x, int y);
  bool mouseLeftUp(gdioutput &gdi, int x, int y);
  bool mouseLeftDblClick(gdioutput &gdi, int x, int y);

  bool editCell(gdioutput &gdi, int row, int col);

  bool character(gdioutput &gdi, int charCode, int repeat);
	void sort(int col);
  void filter(int col, const string &filt, bool forceFilter=false);

	int addColumn(const string &Title, int width, bool isnum);
	void addRow(int rowId);
	void set(int column, oBase &owner, int id, const string &data, 
           bool canEdit=true, CellType type=cellEdit);

  //Reload a row from data
  void reloadRow(int rowId);

  bool UpDown(gdioutput &gdi, int direction);
  bool tabFocus(gdioutput &gdi, int direction);
  bool enter(gdioutput &gdi);
  void escape(gdioutput &gdi);
  bool inputChange(gdioutput &gdi, HWND hEdit);
  void resetColumns();
  void update();

	Table(oEvent *oe_, int rowHeight, 
        const string &name, const string &tname);
	~Table(void);
};


enum {TID_CLASSNAME, TID_COURSE, TID_NUM, TID_ID, TID_MODIFIED,
TID_RUNNER, TID_CLUB, TID_START,
TID_FINISH, TID_STATUS, TID_RUNNINGTIME, TID_PLACE, 
TID_CARD, TID_TEAM, TID_LEG, TID_CONTROL, TID_CODES};
